// VOIDSTAR - Goofing around, Oct. 2024 (approach borrowed from DestinyHunter)
#include <stdio.h>   // printf
#include <stdlib.h>  // srand
//#include <time.h>    // time

#include <core.h>
#include <divmod4_table.h>  //< make sure these tables have enough entries for whatever resolution you're working at

#define MAX_MAP_ROWS 80
//#define MAX_MAP_TOP 4
//#define MAX_MAP_BOTTOM 96
        
#define MAX_MAP_COLS 160
//#define MAX_MAP_COLSML 4
//#define MAX_MAP_COLSMR 184

#define HALFRAND 0x3FFFU

#define DIR_N     0
#define DIR_NE    1
#define DIR_E     2
#define DIR_SE    3
#define DIR_S     4
#define DIR_SW    5
#define DIR_W     6
#define DIR_NW    7

// copy of the background initial image (sort of a double-buffer of what to restore to
// when no color is assigned)  [ this could also be somewhere in the $A000 region ]
static unsigned char g_main_map[MAX_MAP_ROWS][MAX_MAP_COLS];

// to keep track of what rendering state each pixel is at
static unsigned char or_equal_modifier[] = {0xC0,0x30,0x0C,0x03};
static unsigned char cell_state[MAX_MAP_ROWS][40];  // Y/X, 40*(4 half-nibbles per cell) = 160 columns

typedef struct 
{
  unsigned char offset_x;  // adjust to "int" if using >255 width screen resolution
  unsigned char offset_y;
  unsigned char color;
} Location_to_draw;

//#define MAX_LOCATIONS_TO_DRAW 600  //< Obsolete, now using entire 8KB bank (~2700 max locations)
static Location_to_draw* locations_to_draw = (void*)0xA000;  // force this data to current active bank
//static Location_to_draw locations_to_draw[MAX_LOCATIONS_TO_DRAW];

static unsigned char num_locations_to_draw = 0;	

// assumes a 320x240 resolution
void write_pixel(unsigned char x, unsigned char y, unsigned char col)
{
    unsigned char n1,n2;
	static unsigned char n3 = 0;
	unsigned int ofs;  // computed memory offets
	
	ofs = (y*320LU) + x;
	
    n1 = (ofs & 0xFF);
    n2 = (ofs >> 8) & 0xFF;
    //n3 = 0;  //(ofs >> 16) & 0x01;  // ignored because we're limiting ourselves to smol-screen (<256 rows)
	
	POKE(0x9F20, n1);
    POKE(0x9F21, n2);
    POKE(0x9F22, n3);
	
	POKE(0x9F23,col);
}

// this is "fast" since we are using a given offset, don't need to
// combine an x/y
void write_pixel_fast(unsigned long ofs, unsigned char col)
{
    unsigned char n1,n2;	
	static unsigned char n3 = 0;
	
	n1 = (ofs & 0xFF);
    n2 = (ofs >> 8) & 0xFF;
    //n3 = (ofs >> 16) & 0x01;
	
	POKE(0x9F20,n1);
    POKE(0x9F21,n2);
    POKE(0x9F22,n3);
	
	POKE(0x9F23,col);
}

void write_pixel_incI(unsigned long ofs, unsigned char col)
{
    unsigned char n1,n2;	
	static unsigned char n3 = 0;
	
	n1 = (ofs & 0xFF);
    n2 = (ofs >> 8) & 0xFF;
    //n3 = (ofs >> 16) & 0x01;  // enable auto increment 1 byte
	
	POKE(0x9F20,n1);
    POKE(0x9F21,n2);
    POKE(0x9F22,n3 | 0x10);
	
	POKE(0x9F23,col);
}

#define WRITE_PIXEL_INC(c) \
  POKE(0x9F23,c);

#define MARK_LOCATION_AS_DRAWN(data_x,data_y)    \
	cell_state[ data_y ][ div4_table[data_x] ] |= or_equal_modifier[mod4_table[data_x]];	

/*
function alternative to the macro defined below: using either one is a compromise
between code-size and runtime performance...
void BUFFER_LOCATION_TO_DRAW(unsigned char data_x, unsigned char data_y, unsigned char a_color)
{
	MARK_LOCATION_AS_DRAWN(data_x, data_y);
	locations_to_draw[num_locations_to_draw].offset_x = data_x;
	locations_to_draw[num_locations_to_draw].offset_y = data_y;
	locations_to_draw[num_locations_to_draw].color = a_color;
	++num_locations_to_draw;
}
*/

#define BUFFER_LOCATION_TO_DRAW(data_x, data_y, a_color) \
	MARK_LOCATION_AS_DRAWN(data_x, data_y); \
	locations_to_draw[num_locations_to_draw].offset_x = data_x; \
	locations_to_draw[num_locations_to_draw].offset_y = data_y; \
	locations_to_draw[num_locations_to_draw].color = a_color; \
	++num_locations_to_draw;

void run_stage()
{
	Location_to_draw* location_to_draw_ptr;	
	
	register unsigned char i;  //< Since this is used in a lot of logic, we keep the name short.  But it should be called: temp_cell_state
	register unsigned char temp_x;  //< temp_x used for walking the cell_state, also used to move icons, and in drawing the icons (just the portion that would fit on the screen)
	register unsigned char temp_y;  //< temp_y used for walking the cell_state, also used to move icons, and in drawing the icons (just the portion that would fit on the screen)

	unsigned char* ptr_map;	
	unsigned char* ptr_cell;	
	unsigned int x_addr;
	
	// ** RESTORE STAGE MAP PORTIONS THAT ARE NO LONGER COVERED ******** 
	{			
		ptr_map = (unsigned char*)(&g_main_map[0][0]);
		ptr_cell = (unsigned char*)(&cell_state[0][0]);			
		
		temp_x = 0;  // keep track of where we are currently processing (horizontal)
		temp_y = 0;  // keep track of where we are currently processing (vertical)

		x_addr = cx16_screen_row_offset[temp_y];

		while (TRUE)  
		{			
			{			
				if (temp_x >= MAX_MAP_COLS)
				{						
					++temp_y;  // go to the next row
					if (temp_y >= MAX_MAP_ROWS) break;  // finally done!
					
					x_addr = cx16_screen_row_offset[temp_y];
					temp_x = 0;
				}					
					
				// 0x11 (3) was drawn this last animation frame (already on the screen)
				//      (2) could be used to let icons linger one extra frame (not used here)
				// 0x01 (1) was previously drawn on -- animation may stay (increase it back to 2) or it will drop back 0 (background)
				// 0x00 (0) no change
					
				i = (*ptr_cell);
					
				// 1 2 3 4 5 6 7 8
				// [1] [2] [3] [4]
										
				if (i == 0)  // confirmed: this absolutely makes a difference!
				{
					// nothing to do, no animation impacts for this group, fast-track to next
					x_addr += 4;
					ptr_map += 4;
					
					temp_x += 4;
				}
				else
				{					
					// 7 6 5 4 3 2 1 0
					// 1 1 0 0 0 0 0 0 			C0
					// 0 0 1 1 1 1 1 1          3F
					// 0 1 0 0 0 0 0 0          40
					if ((i & 0xC0) == 0xC0) { i &= 0x3F; i |= 0x40; }  // reduce cell_1_state from 2 to 1
					else if ((i & 0xC0) == 0x40)
					{
						// draw map at cell_1 location
						write_pixel_fast(x_addr, *ptr_map);
						i &= 0x3F;  //< reduce cell_1_state from 1 to 0
					}
					++x_addr;
					++ptr_map;
					++temp_x;
					
					// 7 6 5 4 3 2 1 0
					// 0 0 1 1 0 0 0 0 			30
					// 1 1 0 0 1 1 1 1          CF
					// 0 0 0 1 0 0 0 0          10						
					if ((i & 0x30) == 0x30) { i &= 0xCF; i |= 0x10; }  // reduce cell_2_state from 2 to 1
					else if ((i & 0x30) == 0x10)
					{
						// draw map at cell_2 location
						write_pixel_fast(x_addr, *ptr_map);
						i &= 0xCF;  //< reduce cell_1_state from 1 to 0
					}
					++x_addr;
					++ptr_map;
					++temp_x;
						
					// 7 6 5 4 3 2 1 0
					// 0 0 0 0 1 1 0 0 			0C
					// 1 1 1 1 0 0 1 1          F3
					// 0 0 0 0 0 1 0 0          04
					if ((i & 0x0C) == 0x0C) { i &= 0xF3; i |= 0x04; }  // reduce cell_2_state from 2 to 1
					else if ((i & 0x0C) == 0x04)
					{
						// draw map at cell_3 location, 
						write_pixel_fast(x_addr, *ptr_map);
						i &= 0xF3;  //< reduce cell_3_state from 1 to 0
					}
					++x_addr;
					++ptr_map;
					++temp_x;
					
					// 7 6 5 4 3 2 1 0
					// 0 0 0 0 0 0 1 1 			03
					// 1 1 1 1 1 1 0 0          FC
					// 0 0 0 0 0 0 0 1          01						
					if ((i & 0x03) == 0x03) { i &= 0xFC; i |= 0x01; }  // reduce cell_2_state from 2 to 1
					else if ((i & 0x03) == 0x01)
					{
						// draw map at cell_4 location
						write_pixel_fast(x_addr, *ptr_map);
						i &= 0xFC;  //< reduce cell_4_state from 1 to 0
					}
					++x_addr;
					++ptr_map;
					++temp_x;

					(*ptr_cell) = i;
				}
					
				++ptr_cell;				
			}
		}
		// ******************************************

		// ** DRAW ALL CELLS MARKED TO BE DRAWN **************************		
	    location_to_draw_ptr =  &(locations_to_draw[num_locations_to_draw-1]);
		while (num_locations_to_draw > 0)
		{		    
			write_pixel(
				location_to_draw_ptr->offset_x,
				location_to_draw_ptr->offset_y,
				location_to_draw_ptr->color
			);	
						
			--num_locations_to_draw;
			--location_to_draw_ptr;
		}
		// ***************************************************************
	}	
}

void enable_screen_80()
{
	__asm__("lda #$80");  // 320x240
	__asm__("clc");
	__asm__("jsr $ff5f");
}

void show_mouse()
{
	__asm__("sec");
	__asm__("jsr $ff5f");  // get current screen mode
	__asm__("lda #1");     
	__asm__("jsr $ff68");  // show default mouse cursor
}

void poll_mouse()
{
	__asm__("LDX #$70");
    __asm__("JSR $ff6b");  // get mouse position in $70/$71 (X) and $72/$73 (Y)
	__asm__("sta $400");   // store button-state into golden RAM
	__asm__("stx $401");   // store mouse-wheel-state into golden RAM
}

void clear_grx_screen(unsigned char color)
{
  unsigned int ofs = 4800U;  // proportional to the number of inline POKE's done below (i.e. loop-unroll)
  
  // enable auto-byte increment, init offset to VERA addr 0
  POKE(0x9F22,0x10);
  POKE(0x9F21,0);
  POKE(0x9F20,0);
    
  while (ofs > 0U)
  {
	// 16 * 4800 = 76,800 == 320x240
    POKE(0x9F23, color);  // 1
	POKE(0x9F23, color);  // 2
	POKE(0x9F23, color);  // 3
	POKE(0x9F23, color);  // 4
	POKE(0x9F23, color);  // 5
	POKE(0x9F23, color);  // 6
	POKE(0x9F23, color);  // 7
	POKE(0x9F23, color);  // 8
    POKE(0x9F23, color);  // 9
	POKE(0x9F23, color);  // 10
	POKE(0x9F23, color);  // 11
	POKE(0x9F23, color);  // 12
	POKE(0x9F23, color);  // 13
	POKE(0x9F23, color);  // 14
	POKE(0x9F23, color);  // 15
	POKE(0x9F23, color);  // 16	
	
	--ofs;
  }  
  
  // turn off auto-increment (and set high-byte to 0, we're not going to use it)
  POKE(0x9F22, 0);
}

void query_timer_4byte()
{
	__asm__("jsr $FFDE");  // get current system clock
	__asm__("stz $600");  // WARNING: There seems to be a bug in cc65 related to copying/assigning unsigned long	
	__asm__("sty $601");
	__asm__("stx $602");     				
	__asm__("sta $603");				
}

// ptr == unsigned char *
// result = unsigned int
#define query_timer_2byte(ptr, result)      \
	__asm__("jsr $FFDE");                   \
	__asm__("stx $602");                    \
	__asm__("sta $603");                    \
	ptr = (unsigned char *)(&result);       \
    POKE(idx_ptr, PEEK(0x0603));            \
    ++idx_ptr;                              \
    POKE(idx_ptr, PEEK(0x0602));

// JSR joystick_scan
#define GAMEPAD_GET       \
    __asm__("LDA #1");    \
	__asm__("JSR $FF56"); \
	__asm__("STX $604");  \
	__asm__("STY $605");  \
	__asm__("STA $606");


unsigned char g_wiggle = 1;  // large value is more wiggle during cruise control
unsigned char g_wiggle1 = 254, g_wiggle2 = 253, g_wiggle3 = 252;

void handle_wiggle(signed char mw)
{
	if ((mw < 0) && (g_wiggle < 16)) ++g_wiggle;
	if ((mw > 0) && (g_wiggle > 0)) --g_wiggle;
	
	g_wiggle1 = 255-g_wiggle;
	g_wiggle2 = g_wiggle1-g_wiggle;
	g_wiggle3 = g_wiggle2-g_wiggle;	
}

int main()
{  
  unsigned int timer1_prior;
  unsigned int timer2_prior;
  unsigned int timer_now;
  unsigned int timer1_delta;  // time to go forward
  unsigned int timer2_delta;  // timer for cruise control
  
  unsigned char x,tx;
  unsigned char y,ty;  
  unsigned char* idx_ptr;
  
  unsigned long ofs;
  unsigned int i;  
  unsigned char ch, c;  // c used for color
  signed char mw;
  
  unsigned char bottom_margin = MAX_MAP_ROWS-8;
  unsigned char right_margin = MAX_MAP_COLS-8;
  
  // set initial starting position of the worm "body" (head-to-tail)
  unsigned char worm_x[8] = {80, 80, 80, 80, 80, 80, 80, 80};
  unsigned char worm_y[8] = {40, 41, 42, 43, 44, 45, 46, 47};
  unsigned char worm_d = 0;  // 0 = N, 1 = NE, 2 = E, 3 = SE, 4 = S, 5 = SW, 6 = W, 7 = NW  
  
  enable_screen_80();
  
  // Keep these the same so that we have 1:1 pixel size (square)  
  POKE(0x9F2A, 0x20);  // X horizontal scaling
  POKE(0x9F2B, 0x20);  // Y vertical scaling    
  
  // set palette index used on border TOP/BOTTOM  (byte 1) (color index 13)
  POKE(0x9F20, 26);    
  POKE(0x9F21, 0xFA);
  POKE(0x9F22, 0x01);
  // -----------
  POKE(0x9F23, 0xFF);  // GREEN | BLUE

  POKE(0x9F20, 27);    // palette index used on TOP/BOTTOM (byte 2)
  POKE(0x9F21, 0xFA);
  POKE(0x9F22, 0x01);
  // -----------
  POKE(0x9F23, 0x07);	  // 00 | RED


  // palette index used on border LEFT/RIGHT sides  (color index 14)
  POKE(0x9F20, 28);    
  POKE(0x9F21, 0xFA);
  POKE(0x9F22, 0x01);
  // -----------
  POKE(0x9F23, 0xFF);  // GREEN | BLUE
	
  POKE(0x9F20, 29);    // palette index used on sides (byte 2)
  POKE(0x9F21, 0xFA);
  POKE(0x9F22, 0x01);
  // -----------
  POKE(0x9F23, 0x07);	  // 00 | RED
	
  query_timer_2byte(idx_ptr, timer1_prior);  
  timer2_prior = timer1_prior;  // set all timers to the same starting point
  
  srand(timer1_prior);  // seed RNG (just borrowing timer1 for it)
       
  clear_grx_screen(0);
    
  // ===============================
  // generate a background (emulate carbon fiber) - this is slow in code, but would be replaced
  // by a BVLOAD in final      
  for (y = 0; y < MAX_MAP_ROWS; ++y)
  {	  
	  
	  ofs = cx16_screen_row_offset[y];
	  c = 20 + (y % 3);  // start with some shade of gray
	  
	  g_main_map[y][0] = c;
	  write_pixel_incI(ofs, c);  // 0th index of this y-row
	  
	  for (x = 1; x < MAX_MAP_COLS; ++x)
	  {
		  ++c;
		  if (c > 22) c = 20;  // cycle through some grays
		  
		  g_main_map[y][x] = c;
		  		 
	  	  //write_pixel(x, y, c);				 
		  WRITE_PIXEL_INC(c);  //< faster
	  }
	  
  }
  
  // draw vertical (left/right) lines
  c = 14;
  for (y = 4; y < MAX_MAP_ROWS-4; ++y)
  {
	  ofs = cx16_screen_row_offset[y];
	  
	  g_main_map[y][4] = c;
	  write_pixel(4, y, c);
	  
	  g_main_map[y][MAX_MAP_COLS-4] = c;
	  write_pixel(MAX_MAP_COLS-4, y, c);
  }
  // draw horizontal (upper/lower) lines
  c = 13;
  for (x = 5; x < MAX_MAP_COLS-4; ++x)
  {		  
	  g_main_map[4][x] = c;
	  write_pixel(x, 4, c);				 	  
	  
	  g_main_map[MAX_MAP_ROWS-5][x] = c;
	  write_pixel(x, MAX_MAP_ROWS-5, c);				 	  
  }
  // ===============================  
    
  printf("\n\n\n\n\n\n\n\n\n\n\nPress LEFT/RIGHT\narrow key to turn");
  
  show_mouse();
  
  while (TRUE)
  {
	poll_mouse();
	c = PEEK(0x0400);
	if ((c & 0x01) == 0x01)
	{		
		tx = PEEK(0x70);
		ty = PEEK(0x72);				
		if (
		  ((tx > 6) && (tx < MAX_MAP_COLS-6))
		  && ((ty > 6) && (ty < MAX_MAP_ROWS-7))
		)
		{
			//i = rand();  // 0..32767		
			//if (i > HALFRAND) c = 14;
			//else c = 13;
			c = 15;
			g_main_map[ty][tx] = c;	
			write_pixel(tx, ty, c);
		}
	}	
	else if ((c & 0x02) == 0x02)
	{		
		tx = PEEK(0x70);
		ty = PEEK(0x72);
		if (
		  ((tx > 6) && (tx < MAX_MAP_COLS-6))
		  && ((ty > 6) && (ty < MAX_MAP_ROWS-7))
		)
		{
			//if ((g_main_map[ty][tx] == 14) || ((g_main_map[ty][tx] == 15)))
		    if (g_main_map[ty][tx] == 15)
			{	    
		      g_main_map[ty][tx] = 0;
		      write_pixel(tx, ty, 0);
			}
		}
	}
	
	mw = PEEK(0x0401);
	if (mw != 0)
	{
		handle_wiggle(mw);
	}	
	
	/*
	++c;
	for (ch = 0; ch < 8; ++ch)
	{
	  BUFFER_LOCATION_TO_DRAW(worm_x[ch], worm_y[ch], c);	  
	}
	// unroll the above loop into the following...
	
	// NOTE: refer to default VERA palette half-way down the following
	// https://github.com/X16Community/x16-docs/blob/master/X16%20Reference%20-%2009%20-%20VERA%20Programmer's%20Reference.md
	*/	
	BUFFER_LOCATION_TO_DRAW(worm_x[0], worm_y[0], 87);  // HEAD
	
	BUFFER_LOCATION_TO_DRAW(worm_x[1], worm_y[1], 38);  // start of the body 
	BUFFER_LOCATION_TO_DRAW(worm_x[2], worm_y[2], 37);
	BUFFER_LOCATION_TO_DRAW(worm_x[3], worm_y[3], 36);
	BUFFER_LOCATION_TO_DRAW(worm_x[4], worm_y[4], 35);
	BUFFER_LOCATION_TO_DRAW(worm_x[5], worm_y[5], 34);
	BUFFER_LOCATION_TO_DRAW(worm_x[6], worm_y[6], 33);
	
	x = worm_x[7];
	y = worm_y[7];
	BUFFER_LOCATION_TO_DRAW(x, y, 32);  // tip of tail
    g_main_map[y][x] = 19;  //< use this to "etch" a trail of the worm	
	  
	/*
    // draw a little wall in the corner (experiment)	 
	for (x = 159,y = 81; x>150 && y>70; --x, --y)
	{
	  BUFFER_LOCATION_TO_DRAW(x, y, 1);	  
	}
	*/
	
	// apply any buffered locations to draw, refresh any portion not
	// being draw on.
	//run_stage();
	
	query_timer_2byte(idx_ptr, timer_now);
	timer1_delta = (timer_now - timer1_prior);
	timer2_delta = (timer_now - timer2_prior);
	
	GET_CX16_KEY(ch); 
	c = 0;  // flag to detect user movement (user-manual movements will auto-increment cruise control engagement timer)
	
	if (ch == 0)
	{
		if (timer1_delta > 10)
		{
		  timer1_prior = timer_now;
		  goto force_forward;
		}
		
		if (timer2_delta > 180)  // invoke "cruise control"...
		{			
		    c = 1;  // set flag that this is "artificial movement" (don't reset cruise control timer)
			
			ch = rand();  // 0..32767, but just keep the lower byte
			if (ch > g_wiggle1) ch = KEY_LEFT_ARROW;
			else if (ch > g_wiggle2) ch = KEY_RIGHT_ARROW;
			else if (ch > g_wiggle3) ch = KEY_UP_ARROW;  // "UP" really means "forward"		
			else {
				ch = 0;
				if (g_wiggle > 1) 
				{
					i = rand();  // 0..32767
					if (i > 32000)  // small chance of getting tired
					{
					  handle_wiggle(1); 
					}
				}
			}
		}
	}	
		
	// arrow key movement	
	if (ch != 0)
	{
		// we don't know yet if they pressed a meaningful key...
		// using default system ISR, we're not checking for multiple keys at the same time.
		// this means "per main loop" they can either TURN LEFT/RIGHT or GO FORWARD
		tx = 255;
		ty = 255;

		if (ch == KEY_RIGHT_ARROW) 
		{
	      if (worm_d == DIR_NW)
		  {
			  worm_d = DIR_N;
		  }
	      else 
		  {
	         ++worm_d;
		  }
		  ch = KEY_UP_ARROW;  // force a go-forward		  
		}
		else if (ch == KEY_LEFT_ARROW) 
		{ 		
	      if (worm_d == DIR_N)
		  {
			  worm_d = DIR_NW;
		  }
	      else 
		  {
	         --worm_d;
		  }
		  ch = KEY_UP_ARROW;  // force a go-forward		  
		}
		
		// are they going forward?
		if	(ch == KEY_UP_ARROW)  // "move forward"
		{
			if (c == 0) timer2_prior = timer_now;
force_forward:			
			// copy the current "head" position
			tx = worm_x[0];
			ty = worm_y[0];
			
			switch (worm_d)
			{
				case DIR_N:
				  --ty;
				  break;
				  
				case DIR_NE:
				  ++tx;
				  --ty;
				  break;
				  
				case DIR_E:
				  ++tx;
				  break;
				  
				case DIR_SE:
				  ++tx;
				  ++ty;
				  break;
				  
				case DIR_S:
				  ++ty;
				  break;
				
				case DIR_SW:
				  --tx;
				  ++ty;
				  break;
				  
				case DIR_W:
				  --tx;
				  break;
				  
				case DIR_NW:
				  --tx;
				  --ty;
				  break;
			}
		}
		else if (ch == KEY_DOWN_ARROW)
		{
			// not currently supporting a "move backwards"
		}
		
	    if ((tx == 255) && (ty == 255))
		{
			// they pushed some other key, no movement change
		}			
		else
		{
			ch = g_main_map[ty][tx];
			
			// probe where we temporary moved-to, to see if we ran into an edge
			if (ch == 15)
			{
				// 50/50 chance of interpreting the "food" as either type of wall
				// its an effect of how the bounce happens that could end up
				// "eating" the food.
				i = rand();
				ch = 13;
				if (i > HALFRAND) ch = 14;
			}
			
			if (ch == 14)  // if ran into a LEFT/RIGHT wall... (using color index 14 as marker)
			{
				if (worm_d == DIR_NE)
				{
					if (ty < 8)
					{
						// go west
						++ty;
						tx -= 2;
						worm_d = DIR_W;
					}
					else
					{
					  // bounce up-left
					  tx -= 2;
					  worm_d = DIR_NW;
					}
				}
				else if (worm_d == DIR_E)
				{
					i = rand();  // going straight east, randomly decide to go up or down
					if (i > HALFRAND)
					{
						if (ty > bottom_margin) goto nevermind3a;  // near bottom edge, let's just go up
nevermind4a:						
						// go down
						++ty;
						tx -= 2;
						worm_d = DIR_SW;
					}					
					else 
					{
						if (ty < 8) goto nevermind4a;  // near top edge, let's just go down
nevermind3a:						
						// go up
						--ty;
						tx -= 2;
						worm_d = DIR_NW;
					}					
				}
				else if (worm_d == DIR_SE)
				{
					if (ty > bottom_margin)  // too close to bottom, just go west
					{
						// go west
						--ty;
						tx -= 2;
						worm_d = DIR_W;
					}
					else
					{
					  // bounce down-left
					  tx -= 2;
					  worm_d = DIR_SW;
					}
				}
				else if (worm_d == DIR_SW)
				{
					if (ty > bottom_margin)  // too close to bottom, just go east
					{
						// go east
						--ty;
						tx += 2;
						worm_d = DIR_E;
					}
					else
					{
					  // bounce down-right
					  tx += 2;
					  worm_d = DIR_SE;
					}
				}
				else if (worm_d == DIR_W)  // headed straight west, randomly pick to bounce UP or DOWN
				{
					i = rand();
					if (i > HALFRAND)
					{
						if (ty > bottom_margin) goto nevermind8a;  // too close to the bottom
nevermind7a:				 		
						// go down and right
						++ty;
						tx += 2;
						worm_d = DIR_SE;
					}					
					else 
					{
						if (ty < 8) goto nevermind7a;  // too close to the top
nevermind8a:						
						// go up and right
						--ty;
						tx += 2;
						worm_d = DIR_NE;
					}					
				}
				else if (worm_d == DIR_NW)
				{
					if (ty < 8)  // too close to the top, just go east
					{
						// go east
						++ty;
						tx += 2;
						worm_d = DIR_E;
					}
					else
					{
					  // bounce up-right
					  tx += 2;
					  worm_d = DIR_NE;
					}
				}						
			}
			
// ******************************************			
// ******************************************
// ******************************************
			else if (ch == 13)  // if ran into a NORTH/SOUTH wall... (using color index 13 as marker)
			{
				if (worm_d == DIR_N)  // going straight north, randomly decide to bounce left/right
				{
					i = rand();
					if (i > HALFRAND)
					{
nevermind1b:				
                        if (tx > right_margin) goto nevermind2b;							
						// go right
						++tx;
						ty += 2;
						worm_d = DIR_SE;
					}					
					else 
					{
						if (tx < 8) goto nevermind1b;
nevermind2b:						
						// go left
						--tx;
						ty += 2;
						worm_d = DIR_SW;
					}
				}
				else if (worm_d == DIR_NE)
				{
					if (tx > right_margin)
					{
						// go down
						--tx;
						ty += 2;
						worm_d = DIR_S;
					}
					else
					{
					  // bounce down-right
					  ty += 2;
					  worm_d = DIR_SE;
					}
				}
				else if (worm_d == DIR_SE)
				{
					if (tx > right_margin)
					{
						// go north
						--tx;
						ty -= 2;
						worm_d = DIR_N;
					}
					else
					{
					  // bounce up-right					  
					  ty -= 2;
					  worm_d = DIR_NE;
					}
				}
				else if (worm_d == DIR_S)  // going straight south, randomly bounce back left/right to north
				{
					i = rand();
					if (i > HALFRAND)
					{
nevermind5b:				
                        if (tx > right_margin) goto nevermind6b;  // too close to edge							
						// go north-right
						++tx;
						ty -= 2;
						worm_d = DIR_NE;
					}					
					else 
					{
						if (tx < 8) goto nevermind5b;  // too close to edge
nevermind6b:						
						// go north-left
						--tx;
						ty -= 2;
						worm_d = DIR_NW;
					}
				}
				else if (worm_d == DIR_SW)
				{
					if (tx < 8)  // too close to edge
					{
						// go north
						++tx;
						ty -= 2;
						worm_d = DIR_N;
					}
					else
					{
					  // bounce up-left
					  ty -= 2;
					  worm_d = DIR_NW;
					}
				}
				else if (worm_d == DIR_NW)
				{
					if (tx < 8)  // too close to edge
					{
						// go down
						++tx;
						ty += 2;
						worm_d = DIR_S;
					}
					else
					{
					  // bounce down-left
					  ty += 2;
					  worm_d = DIR_SW;
					}
				}
			}
			
			
			/*
			for (ch = 7; ch > 0; --ch)
			{
				worm_x[ch] = worm_x[ch-1];
				worm_y[ch] = worm_y[ch-1];
			}
			// unroll the loop above as follows...
			*/
			
			// shift the "body" of the worm
			worm_x[7] = worm_x[6];
			worm_y[7] = worm_y[6];
			
			worm_x[6] = worm_x[5];
			worm_y[6] = worm_y[5];
			
			worm_x[5] = worm_x[4];
			worm_y[5] = worm_y[4];
			
			worm_x[4] = worm_x[3];
			worm_y[4] = worm_y[3];
			
			worm_x[3] = worm_x[2];
			worm_y[3] = worm_y[2];
			
			worm_x[2] = worm_x[1];
			worm_y[2] = worm_y[1];
			
			worm_x[1] = worm_x[0];
			worm_y[1] = worm_y[0];

			// set the new "head" position
			worm_x[0] = tx;
			worm_y[0] = ty;
			
			run_stage();
		}
	}		
	else
	{
		// no key was pressed by user
	}
	
  }
  
  return 0;
}